import spacy
import subprocess
import sys

def download_spacy_model(model_name):
    """
    Download the specified SpaCy model if it's not already installed.
    """
    try:
        spacy.load(model_name)
    except OSError:
        print(f"Model '{model_name}' not found. Downloading it now...")
        subprocess.run([sys.executable, "-m", "spacy", "download", model_name], check=True)
        print(f"Model '{model_name}' downloaded successfully.")
        spacy.load(model_name)

# Model name
model_name = "en_core_web_sm"

# Download the model if necessary
download_spacy_model(model_name)

# Load the SpaCy model
nlp = spacy.load(model_name)

# Example text
text = """
Slovakia’s prime minister, Robert Fico, is out of immediate danger but remains in intensive care four days after he was shot by a gunman, the country’s deputy prime minister has said.

“He has emerged from the immediate threat to his life, but his condition remains serious and he requires intensive care,” Robert Kaliňák, Fico’s closest political ally, told reporters.

“We can consider his condition stable with a positive prognosis,” Kaliňák said outside the hospital in the central city of Banská Bystrica, where Fico is being treated. He added: “We all feel a bit more relaxed now.”

Kaliňák said Fico would stay in the Banska Bystrica hospital for the time being, adding that his condition was still too serious to allow him to be transferred to a hospital in the capital, Bratislava.

Fico, 59, was shot on Wednesday while walking to greet supporters after a government meeting in the central mining town of Handlová.

Kaliňák said earlier that Fico had suffered four gunshot wounds, two light, one moderate and one serious.
"""

# Process the text with SpaCy
doc = nlp(text)

# Extract named entities
entities = [(ent.text, ent.label_) for ent in doc.ents]

print("Named entities:",  entities)

# Print the named entities
for entity in entities:
    print(f"Entity: {entity[0]}, Label: {entity[1]}")

# Create graph
import networkx as nx
import matplotlib.pyplot as plt

# Initialize graph
G = nx.Graph()

# Add nodes for entities
for entity, label in entities:
    G.add_node(entity, label=label)

# Extract relationships
relationships = []
for sent in doc.sents:
    if "works at" in sent.text or "colleague of" in sent.text or "located in" in sent.text:
        entities_in_sent = [ent.text for ent in sent.ents]
        if len(entities_in_sent) >= 2:
            relationships.append((entities_in_sent[0], entities_in_sent[1]))

# Add edges for relationships
for relationship in relationships:
    G.add_edge(relationship[0], relationship[1])
#
# # Visualize the graph
# pos = nx.spring_layout(G)
# nx.draw(G, pos, with_labels=True, node_color="lightblue", edge_color="gray", node_size=2000, font_size=10)
# plt.show()


from typing import Union, Dict
import re

from pydantic import BaseModel

from cognee.modules.search.llm.extraction.categorize_relevant_category import categorize_relevant_category

""" Search categories in the graph and return their summary attributes. """

from cognee.shared.data_models import GraphDBType, DefaultContentPrediction
import networkx as nx

from cognee.infrastructure.databases.graph.config import get_graph_config
graph_config = get_graph_config()
from cognee.infrastructure.databases.vector.config import get_vectordb_config
vector_config = get_vectordb_config()

def strip_exact_regex(s, substring):
    # Escaping substring to be used in a regex pattern
    pattern = re.escape(substring)
    # Regex to match the exact substring at the start and end
    return re.sub(f"^{pattern}|{pattern}$", "", s)


class DefaultResponseModel(BaseModel):
    document_id: str

async def search_categories(query:str, graph: Union[nx.Graph, any], query_label: str=None, infrastructure_config: Dict=None):
    """
    Filter nodes in the graph that contain the specified label and return their summary attributes.
    This function supports both NetworkX graphs and Neo4j graph databases.

    Parameters:
    - graph (Union[nx.Graph, AsyncSession]): The graph object or Neo4j session.
    - query_label (str): The label to filter nodes by.
    - infrastructure_config (Dict): Configuration that includes the graph engine type.

    Returns:
    - Union[Dict, List[Dict]]: For NetworkX, returns a dictionary where keys are node identifiers,
      and values are their 'content_labels' attributes. For Neo4j, returns a list of dictionaries,
      each representing a node with 'nodeId' and 'summary'.
    """
    # Determine which client is in use based on the configuration
    from cognee.infrastructure import infrastructure_config
    if graph_config.graph_engine == GraphDBType.NETWORKX:

        categories_and_ids = [
            {'document_id': strip_exact_regex(_, "DATA_SUMMARY__"), 'Summary': data['summary']}
            for _, data in graph.nodes(data=True)
            if 'summary' in data
        ]
        connected_nodes = []
        for id in categories_and_ids:
            print("id", id)
            connected_nodes.append(list(graph.neighbors(id['document_id'])))
        check_relevant_category = await categorize_relevant_category(query, categories_and_ids, response_model=DefaultResponseModel )
        connected_nodes = list(graph.neighbors(check_relevant_category['document_id']))
        descriptions = {node: graph.nodes[node].get('description', 'No desc available') for node in connected_nodes}
        return descriptions

    elif graph_config.graph_engine == GraphDBType.NEO4J:
        # Logic for Neo4j
        cypher_query = """
        MATCH (n)
        WHERE $label IN labels(n) AND EXISTS(n.summary)
        RETURN id(n) AS nodeId, n.summary AS summary
        """
        result = await graph.run(cypher_query, label=query_label)
        nodes_summary = [{"nodeId": record["nodeId"], "summary": record["summary"]} for record in await result.list()]
        return nodes_summary

    else:
        raise ValueError("Unsupported graph engine type.")



from cognee.infrastructure.InfrastructureConfig import infrastructure_config
from cognee.infrastructure.databases.relational.config import get_relationaldb_config

config = get_relationaldb_config()

def update_task_status(data_id: str, status: str):
    db_engine = config.db_engine
    db_engine.insert_data("cognee_task_status", [dict(data_id = data_id, status = status)])



from cognee.infrastructure.databases.graph.get_graph_client import get_graph_client
from cognee.infrastructure.databases.graph.config import get_graph_config
graph_config = get_graph_config()
from cognee.infrastructure.databases.vector.config import get_vectordb_config
vector_config = get_vectordb_config()

async def search_similarity(query: str, graph):
    graph_db_type = graph_config.graph_engine

    graph_client = await get_graph_client(graph_db_type)

    layer_nodes = await graph_client.get_layer_nodes()

    unique_layer_uuids = set(node["layer_id"] for node in layer_nodes)
    print("unique_layer_uuids", unique_layer_uuids)


    graph_nodes = []

    for layer_id in unique_layer_uuids:
        vector_engine = vector_config.vector_engine

        results = await vector_engine.search(layer_id, query_text = query, limit = 10)
        print("results", results)
        print("len_rs", len(results))

        if len(results) > 0:
            graph_nodes.extend([
                dict(
                    layer_id = result.payload["references"]["cognitive_layer"],
                    node_id = result.payload["references"]["node_id"],
                    score = result.score,
                ) for result in results if result.score > 0.3
            ])

    if len(graph_nodes) == 0:
        return []


    return graph_nodes



    # for graph_node_data in graph_nodes:
    #     if graph_node_data['score'] >0.8:
    #         graph_node = await graph_client.extract_node(graph_node_data["node_id"])
    #
    #         if "chunk_collection" not in graph_node and "chunk_id" not in graph_node:
    #             continue
    #
    #         vector_point = await vector_engine.retrieve(
    #             graph_node["chunk_collection"],
    #             graph_node["chunk_id"],
    #         )
    #
    #         print("vector_point", vector_point.payload["text"])
    #
    #         relevant_context.append(vector_point.payload["text"])
    #
    # print(relevant_context)
    #
    # return deduplicate(relevant_context)



import os
from functools import lru_cache
from pydantic_settings import BaseSettings, SettingsConfigDict
from cognee.infrastructure.databases.relational.config import get_relationaldb_config
from cognee.infrastructure.databases.vector.embeddings.config import get_embedding_config
from .create_vector_engine import create_vector_engine

embeddings_config = get_embedding_config()
relational_config = get_relationaldb_config()

lancedb_path = os.path.join(relational_config.database_directory_path, "cognee.lancedb")

class VectorConfig(BaseSettings):
    vector_db_url: str = lancedb_path
    vector_db_key: str = ""
    vector_engine_provider: str = "lancedb"
    vector_engine: object = create_vector_engine(
        {
            "vector_db_key": None,
            "vector_db_url": lancedb_path,
            "vector_db_provider": "lancedb",
        },
        embeddings_config.embedding_engine,
    )

    model_config = SettingsConfigDict(env_file = ".env", extra = "allow")

    def create_engine(self):
        if self.vector_engine_provider == "lancedb":
            self.vector_db_url = lancedb_path

        self.vector_engine = create_vector_engine(
            get_vectordb_config().to_dict(),
            embeddings_config.embedding_engine,
        )

    def to_dict(self) -> dict:
        return {
            "vector_db_url": self.vector_db_url,
            "vector_db_key": self.vector_db_key,
            "vector_db_provider": self.vector_engine_provider,
        }

@lru_cache
def get_vectordb_config():
    return VectorConfig()



"""Factory function to get the appropriate graph client based on the graph type."""

from cognee.shared.data_models import GraphDBType
from .config import get_graph_config
from .graph_db_interface import GraphDBInterface
from .networkx.adapter import NetworkXAdapter
config = get_graph_config()


async def get_graph_client(graph_type: GraphDBType, graph_file_name: str = None) -> GraphDBInterface :
    """Factory function to get the appropriate graph client based on the graph type."""

    if graph_type == GraphDBType.NEO4J:
        try:
            from .neo4j_driver.adapter import Neo4jAdapter

            return Neo4jAdapter(
                graph_database_url = config.graph_database_url,
                graph_database_username = config.graph_database_username,
                graph_database_password = config.graph_database_password
            )
        except:
            pass

    elif graph_type == GraphDBType.FALKORDB:
        try:
            from .falkordb.adapter import FalcorDBAdapter

            return FalcorDBAdapter(
                graph_database_url = config.graph_database_url,
                graph_database_username = config.graph_database_username,
                graph_database_password = config.graph_database_password,
                graph_database_port = config.graph_database_port
            )
        except:
            pass
    graph_client = NetworkXAdapter(filename = config.graph_file_path)
    if (graph_client.graph is None):
        await graph_client.load_graph_from_file()

    return graph_client



import asyncio
from typing import List, Optional
from openai import AsyncOpenAI
from fastembed import TextEmbedding

from cognee.root_dir import get_absolute_path
from cognee.infrastructure.databases.vector.embeddings.EmbeddingEngine import EmbeddingEngine
from litellm import aembedding
import litellm

litellm.set_verbose = True

class DefaultEmbeddingEngine(EmbeddingEngine):
    embedding_model: str
    embedding_dimensions: int
    def __init__(
        self,
        embedding_model: Optional[str],
        embedding_dimensions: Optional[int],
    ):
        self.embedding_model = embedding_model
        self.embedding_dimensions = embedding_dimensions

    async def embed_text(self, text: List[str]) -> List[float]:
        embedding_model = TextEmbedding(model_name = self.embedding_model, cache_dir = get_absolute_path("cache/embeddings"))
        embeddings_list = list(map(lambda embedding: embedding.tolist(), embedding_model.embed(text)))

        return embeddings_list

    def get_vector_size(self) -> int:
        return self.embedding_dimensions


class LiteLLMEmbeddingEngine(EmbeddingEngine):
    embedding_model: str
    embedding_dimensions: int
    def __init__(
        self,
        embedding_model: Optional[str],
        embedding_dimensions: Optional[int],
    ):
        self.embedding_model = embedding_model
        self.embedding_dimensions = embedding_dimensions
    import asyncio
    from typing import List

    async def embed_text(self, text: List[str]) -> List[List[float]]:
        async def get_embedding(text_):
            response = await aembedding(self.embedding_model, input=text_)
            return response.data[0]['embedding']

        tasks = [get_embedding(text_) for text_ in text]
        result = await asyncio.gather(*tasks)
        return result
    def get_vector_size(self) -> int:
        return self.embedding_dimensions


if __name__ == "__main__":
    async def gg():
        openai_embedding_engine = LiteLLMEmbeddingEngine()
        # print(openai_embedding_engine.embed_text(["Hello, how are you?"]))
        # print(openai_embedding_engine.get_vector_size())
        # default_embedding_engine = DefaultEmbeddingEngine()
        sds = await openai_embedding_engine.embed_text(["Hello, sadasdas are you?"])
        print(sds)
        # print(default_embedding_engine.get_vector_size())

    asyncio.run(gg())




from __future__ import annotations
import re

from cognee.infrastructure.data.chunking.DefaultChunkEngine import DefaultChunkEngine
from cognee.shared.data_models import ChunkStrategy



class LangchainChunkEngine():
    @staticmethod
    def chunk_data(
        chunk_strategy = None,
        source_data = None,
        chunk_size = None,
        chunk_overlap = None,
    ):
        """
        Chunk data based on the specified strategy.

        Parameters:
        - chunk_strategy: The strategy to use for chunking.
        - source_data: The data to be chunked.
        - chunk_size: The size of each chunk.
        - chunk_overlap: The overlap between chunks.

        Returns:
        - The chunked data.
        """

        if chunk_strategy == ChunkStrategy.CODE:
            chunked_data = LangchainChunkEngine.chunk_data_by_code(source_data,chunk_size, chunk_overlap)
        else:
            chunked_data = DefaultChunkEngine.chunk_data_by_paragraph(source_data,chunk_size, chunk_overlap)
        return chunked_data

    @staticmethod
    def chunk_data_by_code(data_chunks, chunk_size, chunk_overlap, language=None):
        from langchain_text_splitters import (
            Language,
            RecursiveCharacterTextSplitter,
        )
        if language is None:
            language = Language.PYTHON
        python_splitter = RecursiveCharacterTextSplitter.from_language(
            language=language, chunk_size=chunk_size, chunk_overlap=chunk_overlap
        )
        code_chunks = python_splitter.create_documents([data_chunks])

        only_content = [chunk.page_content for chunk in code_chunks]

        return only_content




"""Configuration for cognee - cognitive architecture framework."""
import logging
import os
import configparser
import uuid
from typing import Optional, Dict, Any
from dataclasses import dataclass, field
from pathlib import Path
from dotenv import load_dotenv
from cognee.root_dir import get_absolute_path
from cognee.shared.data_models import ChunkStrategy, DefaultGraphModel
logging.basicConfig(level=logging.DEBUG)
def load_dontenv():
    base_dir = Path(__file__).resolve().parent.parent
    # Load the .env file from the base directory
    dotenv_path = base_dir / ".env"
    load_dotenv(dotenv_path=dotenv_path, override = True)

try:
    load_dontenv()
except:
    pass

@dataclass
class Config:
    """ Configuration for cognee - cognitive architecture framework. """
    cognee_dir: str = field(
        default_factory=lambda: os.getenv("COG_ARCH_DIR", "cognee")
    )
    config_path: str = field(
        default_factory=lambda: os.path.join(
            os.getenv("COG_ARCH_DIR", "cognee"), "config"
        )
    )


    system_root_directory = get_absolute_path(".cognee_system")
    logging.info("system_root_directory: %s", system_root_directory)
    data_root_directory = os.getenv("DATA_PATH", get_absolute_path(".data"))

    vectordb: str = os.getenv("VECTORDB", "weaviate")

    qdrant_path: str = os.getenv("QDRANT_PATH", None)
    qdrant_url: str = os.getenv("QDRANT_URL", None)
    qdrant_api_key: str = os.getenv("QDRANT_API_KEY", None)



    graph_filename = os.getenv("GRAPH_NAME", "cognee_graph.pkl")

    # Model parameters
    llm_provider: str = os.getenv("LLM_PROVIDER", "openai") #openai, or custom or ollama
    llm_model: str = os.getenv("LLM_MODEL", "gpt-4")
    llm_api_key: str = os.getenv("LLM_API_KEY", os.getenv("OPENAI_API_KEY"))
    llm_endpoint: str = os.getenv("LLM_ENDPOINT", None)

    # custom_model: str = os.getenv("CUSTOM_LLM_MODEL", "llama3-70b-8192") #"mistralai/Mixtral-8x7B-Instruct-v0.1"
    # custom_endpoint: str = os.getenv("CUSTOM_ENDPOINT", "https://api.endpoints.anyscale.com/v1") #"https://api.endpoints.anyscale.com/v1" # pass claude endpoint
    # custom_key: Optional[str] = os.getenv("CUSTOM_LLM_API_KEY")
    # ollama_endpoint: str = os.getenv("CUSTOM_OLLAMA_ENDPOINT", "http://localhost:11434/v1") #"http://localhost:11434/v1"
    # ollama_key: Optional[str] = "ollama"
    # ollama_model: str = os.getenv("CUSTOM_OLLAMA_MODEL", "mistral:instruct") #"mistral:instruct"
    # openai_model: str = os.getenv("OPENAI_MODEL", "gpt-4o" ) #"gpt-4o"
    # model_endpoint: str = "openai"
    # llm_api_key: Optional[str] = os.getenv("OPENAI_API_KEY")
    openai_temperature: float = float(os.getenv("OPENAI_TEMPERATURE", 0.0))
    # openai_embedding_model = "text-embedding-3-large"
    # openai_embedding_dimensions = 3072
    # litellm_embedding_model = "text-embedding-3-large"
    # litellm_embedding_dimensions = 3072

    graphistry_username = os.getenv("GRAPHISTRY_USERNAME")
    graphistry_password = os.getenv("GRAPHISTRY_PASSWORD")

    # Embedding parameters
    embedding_model: str = "BAAI/bge-large-en-v1.5"
    embedding_dimensions: int = 1024
    connect_documents: bool = False

    # Database parameters
    graph_database_provider: str = os.getenv("GRAPH_DB_PROVIDER", "NETWORKX")
    graph_topology:str = DefaultGraphModel
    cognitive_layers_limit: int = 2

    from cognee.shared.data_models import MonitoringTool

    # Monitoring tool
    monitoring_tool: str = os.getenv("MONITORING_TOOL", MonitoringTool.LANGFUSE)

    weaviate_url: str = os.getenv("WEAVIATE_URL")
    weaviate_api_key: str = os.getenv("WEAVIATE_API_KEY")

    # Model parameters and configuration for interlayer scoring
    intra_layer_score_treshold: float = 0.98


    # Client ID
    anon_clientid: Optional[str] = field(default_factory=lambda: uuid.uuid4().hex)

    #Chunking parameters
    # chunk_size: int = 1500
    # chunk_overlap: int = 0
    # chunk_strategy: str = ChunkStrategy.PARAGRAPH

    def load(self):
        """Loads the configuration from a file or environment variables."""
        try:
            load_dontenv()
        except:
            pass
        config = configparser.ConfigParser()
        config.read(self.config_path)

        # Override with environment variables if they exist
        for attr in self.__annotations__:
            env_value = os.getenv(attr.upper())
            if env_value is not None:
                setattr(self, attr, env_value)

        # Load from config file
        if config.sections():
            for section in config.sections():
                for key, value in config.items(section):
                    if hasattr(self, key):
                        setattr(self, key, value)

    def save(self):
        """Saves the current configuration to a file."""
        config = configparser.ConfigParser()

        # Save the current settings to the config file
        for attr, value in self.__dict__.items():
            section, option = attr.split("_", 1)
            if not config.has_section(section):
                config.add_section(section)
            config.set(section, option, str(value))

        with open(self.config_path, "w") as configfile:
            config.write(configfile)

    def to_dict(self) -> Dict[str, Any]:
        """Returns a dictionary representation of the configuration."""
        return {attr: getattr(self, attr) for attr in self.__annotations__}

    @classmethod
    def from_dict(cls, config_dict: Dict[str, Any]) -> "Config":
        """Creates a Config instance from a dictionary."""
        config = cls()
        for attr, value in config_dict.items():
            if hasattr(config, attr):
                setattr(config, attr, value)
        return config


from cognee.base_config import get_base_config
from cognee.infrastructure.files.storage import LocalStorage
from cognee.infrastructure import infrastructure_config
from cognee.infrastructure.databases.graph.get_graph_client import get_graph_client
base_config =get_base_config()
from cognee.infrastructure.databases.graph.config import get_graph_config
from cognee.infrastructure.databases.vector.config import get_vectordb_config
graph_config = get_graph_config()
vector_config = get_vectordb_config()

class prune():
    @staticmethod
    async def prune_data():
        data_root_directory = base_config.data_root_directory
        LocalStorage.remove_all(data_root_directory)

    @staticmethod
    async def prune_system(graph = True, vector = True):
        if graph:
            graph_client = await get_graph_client(graph_config.graph_engine)
            await graph_client.delete_graph()

        if vector:
            vector_client = vector_config.vector_engine
            await vector_client.prune()


if __name__ == "__main__":
    import asyncio
    async def main():
        await prune.prune_data()
        await prune.prune_system()


    asyncio.run(main())



from os import path
import logging
from pathlib import Path
logging.basicConfig(level=logging.DEBUG)
# ROOT_DIR = path.dirname(path.abspath(__file__))
#
# logging.debug("ROOT_DIR: ", ROOT_DIR)
#
# def get_absolute_path(path_from_root: str) -> str:
#     logging.debug("abspath: ", path.abspath(path.join(ROOT_DIR, path_from_root)))
#
#
#     return path.abspath(path.join(ROOT_DIR, path_from_root))
ROOT_DIR = Path(__file__).resolve().parent

logging.basicConfig(level=logging.DEBUG)
logging.debug("ROOT_DIR: %s", ROOT_DIR)

def get_absolute_path(path_from_root: str) -> str:
    absolute_path = ROOT_DIR / path_from_root
    logging.debug("abspath: %s", absolute_path.resolve())
    return str(absolute_path.resolve())



import logging
import os

logging.basicConfig(level=logging.DEBUG)
async def  main():
    from os import path
    import pathlib
    import cognee
    logging.basicConfig(level=logging.DEBUG)

    # print("Working dir: ", str(pathlib.Path(__file__).parent))
    # data_directory_path = str(pathlib.Path(path.join(pathlib.Path(__file__).parent, "../../.data")).resolve())
    # print("Data dir: ", data_directory_path)
    # cognee.config.data_root_directory(data_directory_path)
    #
    # cognee_directory_path = str(pathlib.Path(path.join(pathlib.Path(__file__).parent, "../../.cognee_system")).resolve())
    # print("System dir: ", cognee_directory_path)
    # cognee.config.system_root_directory(cognee_directory_path)

    logging.debug("CURRENCT CWD: %s", pathlib.Path(__file__).parent)
    logging.debug("CURRENCT CWD: %s", os.getcwd())


    dataset_name = "cs_explanations"

    explanation_file_path = os.path.join(pathlib.Path(__file__).parent, "test_data/Natural_language_processing.txt")
    await cognee.add([explanation_file_path], dataset_name)
    
    # dataset_name = "short_stories"
    # # data_directory_path is defined above
    # await cognee.add("data://" + "/Users/runner/work/cognee/cognee/./cognee/tests", dataset_name)

    text_1 = """A quantum computer is a computer that takes advantage of quantum mechanical phenomena.
    At small scales, physical matter exhibits properties of both particles and waves, and quantum computing leverages this behavior, specifically quantum superposition and entanglement, using specialized hardware that supports the preparation and manipulation of quantum states.
    Classical physics cannot explain the operation of these quantum devices, and a scalable quantum computer could perform some calculations exponentially faster (with respect to input size scaling) than any modern "classical" computer. In particular, a large-scale quantum computer could break widely used encryption schemes and aid physicists in performing physical simulations; however, the current state of the technology is largely experimental and impractical, with several obstacles to useful applications. Moreover, scalable quantum computers do not hold promise for many practical tasks, and for many important tasks quantum speedups are proven impossible.
    The basic unit of information in quantum computing is the qubit, similar to the bit in traditional digital electronics. Unlike a classical bit, a qubit can exist in a superposition of its two "basis" states. When measuring a qubit, the result is a probabilistic output of a classical bit, therefore making quantum computers nondeterministic in general. If a quantum computer manipulates the qubit in a particular way, wave interference effects can amplify the desired measurement results. The design of quantum algorithms involves creating procedures that allow a quantum computer to perform calculations efficiently and quickly.
    Physically engineering high-quality qubits has proven challenging. If a physical qubit is not sufficiently isolated from its environment, it suffers from quantum decoherence, introducing noise into calculations. Paradoxically, perfectly isolating qubits is also undesirable because quantum computations typically need to initialize qubits, perform controlled qubit interactions, and measure the resulting quantum states. Each of those operations introduces errors and suffers from noise, and such inaccuracies accumulate.
    In principle, a non-quantum (classical) computer can solve the same computational problems as a quantum computer, given enough time. Quantum advantage comes in the form of time complexity rather than computability, and quantum complexity theory shows that some quantum algorithms for carefully selected tasks require exponentially fewer computational steps than the best known non-quantum algorithms. Such tasks can in theory be solved on a large-scale quantum computer whereas classical computers would not finish computations in any reasonable amount of time. However, quantum speedup is not universal or even typical across computational tasks, since basic tasks such as sorting are proven to not allow any asymptotic quantum speedup. Claims of quantum supremacy have drawn significant attention to the discipline, but are demonstrated on contrived tasks, while near-term practical use cases remain limited.
    """

    text_2 = """A large language model (LLM) is a language model notable for its ability to achieve general-purpose language generation and other natural language processing tasks such as classification. LLMs acquire these abilities by learning statistical relationships from text documents during a computationally intensive self-supervised and semi-supervised training process. LLMs can be used for text generation, a form of generative AI, by taking an input text and repeatedly predicting the next token or word.
    LLMs are artificial neural networks. The largest and most capable, as of March 2024, are built with a decoder-only transformer-based architecture while some recent implementations are based on other architectures, such as recurrent neural network variants and Mamba (a state space model).
    Up to 2020, fine tuning was the only way a model could be adapted to be able to accomplish specific tasks. Larger sized models, such as GPT-3, however, can be prompt-engineered to achieve similar results.[6] They are thought to acquire knowledge about syntax, semantics and "ontology" inherent in human language corpora, but also inaccuracies and biases present in the corpora.
    Some notable LLMs are OpenAI's GPT series of models (e.g., GPT-3.5 and GPT-4, used in ChatGPT and Microsoft Copilot), Google's PaLM and Gemini (the latter of which is currently used in the chatbot of the same name), xAI's Grok, Meta's LLaMA family of open-source models, Anthropic's Claude models, Mistral AI's open source models, and Databricks' open source DBRX.
    """

    #
    # dataset_name = "cs_explanations"
    # await cognee.add(
    #     [
    #         text_1,
    #         text_2
    #     ],
    #     dataset_name
    # )

    await cognee.cognify([ "cs_explanations"])


if __name__ == "__main__":
    import asyncio
    asyncio.run(main())






from typing import Union, Dict
import networkx as nx
from cognee.infrastructure import infrastructure_config

from cognee.modules.search.llm.extraction.categorize_relevant_summary import categorize_relevant_summary
from cognee.shared.data_models import GraphDBType, ResponseSummaryModel
from cognee.infrastructure.databases.graph.config import get_graph_config
graph_config = get_graph_config()
from cognee.infrastructure.databases.vector.config import get_vectordb_config
vector_config = get_vectordb_config()
import re

def strip_exact_regex(s, substring):
    # Escaping substring to be used in a regex pattern
    pattern = re.escape(substring)
    # Regex to match the exact substring at the start and end
    return re.sub(f"^{pattern}|{pattern}$", "", s)
async def search_summary( query: str,  graph: Union[nx.Graph, any]) -> Dict[str, str]:
    """
    Filter nodes based on a condition (such as containing 'SUMMARY' in their identifiers) and return their summary attributes.
    Supports both NetworkX graphs and Neo4j graph databases based on the configuration.

    Parameters:
    - graph (Union[nx.Graph, AsyncSession]): The graph object or Neo4j session.
    - query (str): The query string to filter nodes by, e.g., 'SUMMARY'.
    - infrastructure_config (Dict): Configuration that includes the graph engine type.
    - other_param (str, optional): An additional parameter, unused in this implementation but could be for future enhancements.

    Returns:
    - Dict[str, str]: A dictionary where keys are node identifiers containing the query string, and values are their 'summary' attributes.
    """

    if graph_config.graph_engine == GraphDBType.NETWORKX:
        print("graph", graph)
        summaries_and_ids = [
            {'document_id': strip_exact_regex(_, "DATA_SUMMARY__"), 'Summary': data['summary']}
            for _, data in graph.nodes(data=True)
            if 'summary' in data
        ]
        print("summaries_and_ids", summaries_and_ids)
        check_relevant_summary = await categorize_relevant_summary(query, summaries_and_ids, response_model=ResponseSummaryModel)
        print("check_relevant_summary", check_relevant_summary)

        connected_nodes = list(graph.neighbors(check_relevant_summary['document_id']))
        print("connected_nodes", connected_nodes)
        descriptions = {node: graph.nodes[node].get('description', 'No desc available') for node in connected_nodes}
        print("descs", descriptions)
        return descriptions


    elif graph_config.graph_engine == GraphDBType.NEO4J:
        cypher_query = f"""
        MATCH (n)
        WHERE n.id CONTAINS $query AND EXISTS(n.summary)
        RETURN n.id AS nodeId, n.summary AS summary
        """
        results = await graph.run(cypher_query, query=query)
        summary_data = {record["nodeId"]: record["summary"] for record in await results.list()}
        return summary_data

    else:
        raise ValueError("Unsupported graph engine type in the configuration.")



from functools import lru_cache
from pydantic_settings import BaseSettings, SettingsConfigDict

from cognee.root_dir import get_absolute_path

from cognee.shared.data_models import MonitoringTool, DefaultContentPrediction, LabeledContent, SummarizedContent, \
    DefaultCognitiveLayer, DefaultGraphModel, KnowledgeGraph


# Monitoring tool



class CognifyConfig(BaseSettings):
    classification_model: object = DefaultContentPrediction
    summarization_model: object = SummarizedContent
    labeling_model: object = LabeledContent
    cognitive_layer_model: object = DefaultCognitiveLayer
    intra_layer_score_treshold: float = 0.98
    connect_documents: bool = False
    graph_topology: object = DefaultGraphModel
    cognitive_layers_limit: int = 2
    graph_model:object = KnowledgeGraph



    model_config = SettingsConfigDict(env_file = ".env", extra = "allow")

    def to_dict(self) -> dict:
        return {
            "classification_model": self.classification_model,
            "summarization_model": self.summarization_model,
            "labeling_model": self.labeling_model,
            "cognitive_layer_model": self.cognitive_layer_model,
            "intra_layer_score_treshold": self.intra_layer_score_treshold,
            "connect_documents": self.connect_documents,
            "graph_topology": self.graph_topology,
            "cognitive_layers_limit": self.cognitive_layers_limit,
            "graph_model": self.graph_model
        }

@lru_cache
def get_cognify_config():
    return CognifyConfig()


""" Neo4j Adapter for Graph Database"""
import json
import logging
from typing import Optional, Any, List, Dict
from contextlib import asynccontextmanager
from neo4j import AsyncSession
from neo4j import AsyncGraphDatabase
from neo4j.exceptions import Neo4jError
from cognee.infrastructure.databases.graph.graph_db_interface import GraphDBInterface

logger = logging.getLogger("Neo4jAdapter")

class Neo4jAdapter(GraphDBInterface):
    def __init__(
        self,
        graph_database_url: str,
        graph_database_username: str,
        graph_database_password: str,
        driver: Optional[Any] = None,
    ):
        self.driver = driver or AsyncGraphDatabase.driver(
            graph_database_url,
            auth = (graph_database_username, graph_database_password)
        )

    async def close(self) -> None:
        await self.driver.close()

    @asynccontextmanager
    async def get_session(self) -> AsyncSession:
        async with self.driver.session() as session:
            yield session

    async def query(
        self,
        query: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> List[Dict[str, Any]]:
        try:
            async with self.get_session() as session:
                result = await session.run(query, parameters=params)
                data = await result.data()
                await self.close()
                return data
        except Neo4jError as error:
            logger.error("Neo4j query error: %s", error, exc_info = True)
            raise error

    async def graph(self):
        return await self.get_session()

    async def add_node(self, node_id: str, node_properties: Dict[str, Any] = None):
        node_id = node_id.replace(":", "_")

        serialized_properties = self.serialize_properties(node_properties)

        if "name" not in serialized_properties:
            serialized_properties["name"] = node_id
        query = f"""MERGE (node:`{node_id}` {{id: $node_id}})
                ON CREATE SET node += $properties
                RETURN ID(node) AS internal_id, node.id AS nodeId"""

        params = {
            "node_id": node_id,
            "properties": serialized_properties,
        }

        return await self.query(query, params)

    async def add_nodes(self, nodes: list[tuple[str, dict[str, Any]]]) -> None:
        # nodes_data = []

        for node in nodes:
            node_id, node_properties = node
            node_id = node_id.replace(":", "_")

            await self.add_node(
                node_id = node_id,
                node_properties = node_properties,
            )

    async def extract_node_description(self, node_id: str):
        query = """MATCH (n)-[r]->(m)
                    WHERE n.id = $node_id
                    AND NOT m.id CONTAINS 'DefaultGraphModel'
                    RETURN m
                    """

        result = await self.query(query, dict(node_id = node_id))

        descriptions = []

        for node in result:
            # Assuming 'm' is a consistent key in your data structure
            attributes = node.get("m", {})

            # Ensure all required attributes are present
            if all(key in attributes for key in ["id", "layer_id", "description"]):
                descriptions.append({
                    "id": attributes["id"],
                    "layer_id": attributes["layer_id"],
                    "description": attributes["description"],
                })

        return descriptions

    async def get_layer_nodes(self):
        query = """MATCH (node) WHERE node.layer_id IS NOT NULL
        RETURN node"""

        return [result["node"] for result in (await self.query(query))]

    async def extract_node(self, node_id: str):
        query= """
        MATCH(node {id: $node_id})
        RETURN node
        """

        results = [node["node"] for node in (await self.query(query, dict(node_id = node_id)))]

        return results[0] if len(results) > 0 else None

    async def delete_node(self, node_id: str):
        node_id = id.replace(":", "_")

        query = f"MATCH (node:`{node_id}` {{id: $node_id}}) DETACH DELETE n"
        params = { "node_id": node_id }

        return await self.query(query, params)

    async def add_edge(self, from_node: str, to_node: str, relationship_name: str, edge_properties: Optional[Dict[str, Any]] = {}):
        serialized_properties = self.serialize_properties(edge_properties)
        from_node = from_node.replace(":", "_")
        to_node = to_node.replace(":", "_")

        query = f"""MATCH (from_node:`{from_node}`
         {{id: $from_node}}), 
         (to_node:`{to_node}` {{id: $to_node}})
         MERGE (from_node)-[r:`{relationship_name}`]->(to_node)
         SET r += $properties
         RETURN r"""

        params = {
            "from_node": from_node,
            "to_node": to_node,
            "properties": serialized_properties
        }

        return await self.query(query, params)


    async def add_edges(self, edges: list[tuple[str, str, str, dict[str, Any]]]) -> None:
        # edges_data = []

        for edge in edges:
            from_node, to_node, relationship_name, edge_properties = edge
            from_node = from_node.replace(":", "_")
            to_node = to_node.replace(":", "_")

            await self.add_edge(
                from_node = from_node,
                to_node = to_node,
                relationship_name = relationship_name,
                edge_properties = edge_properties
            )


    async def filter_nodes(self, search_criteria):
        query = f"""MATCH (node)
                WHERE node.id CONTAINS '{search_criteria}'
                RETURN node"""


        return await self.query(query)


    async def delete_graph(self):
        query = """MATCH (node)
                DETACH DELETE node;"""

        return await self.query(query)

    def serialize_properties(self, properties = dict()):
        return {
            property_key: json.dumps(property_value)
            if isinstance(property_value, (dict, list))
            else property_value for property_key, property_value in properties.items()
        }



import logging
from typing import List, Dict
from cognee.infrastructure import infrastructure_config
from cognee.modules.topology.extraction.extract_topology import extract_categories
from cognee.modules.cognify.config import get_cognify_config

cognify_config = get_cognify_config()

logger = logging.getLogger(__name__)

async def infer_data_topology(content: str, graph_topology=None):
    if graph_topology is None:
        graph_topology = cognify_config.graph_topology
    try:
        return (await extract_categories(
            content,
            graph_topology
        ))
    except Exception as error:
        logger.error("Error extracting cognitive layers from content: %s", error, exc_info = True)
        raise error



from cognee.shared.data_models import Document
# from cognee.modules.cognify.graph.add_label_nodes import add_label_nodes
from cognee.infrastructure.databases.graph.graph_db_interface import GraphDBInterface

async def add_document_node(graph_client: GraphDBInterface, parent_node_id, document_metadata):
    document_id = f"DOCUMENT__{document_metadata['id']}"

    document = await graph_client.extract_node(document_id)

    if not document:
        document = Document(
            id = document_id,
            title = document_metadata["name"],
            file_path = document_metadata["file_path"],
        ).model_dump()

    document["type"] = "Document"

    await graph_client.add_node(document_id, document)
    print(f"Added document node: {document_id}")

    await graph_client.add_edge(
        parent_node_id,
        document_id,
        "has_document",
        dict(relationship_name = "has_document"),
    )

    #
    # await add_label_nodes(graph_client, document_id, document_metadata["keywords"].split("|"))

    return document_id




from typing import Union, Dict
import re

import networkx as nx
from pydantic import BaseModel

from cognee.modules.search.llm.extraction.categorize_relevant_category import categorize_relevant_category
from cognee.shared.data_models import GraphDBType
from cognee.infrastructure.databases.graph.config import get_graph_config
graph_config = get_graph_config()
from cognee.infrastructure.databases.vector.config import get_vectordb_config
vector_config = get_vectordb_config()

async def search_cypher(query:str, graph: Union[nx.Graph, any]):
    """
    Use a Cypher query to search the graph and return the results.
    """


    from cognee.infrastructure import infrastructure_config
    if graph_config.graph_engine == GraphDBType.NEO4J:
        result = await graph.run(query)
        return result

    else:
        raise ValueError("Unsupported graph engine type.")


""" This module contains the configuration for the graph database. """
import os
from functools import lru_cache
from pydantic_settings import BaseSettings, SettingsConfigDict
from cognee.base_config import get_base_config
from cognee.infrastructure.databases.relational.config import get_relationaldb_config
from cognee.shared.data_models import DefaultGraphModel, GraphDBType
relational_config = get_relationaldb_config()
base_config = get_base_config()

class GraphConfig(BaseSettings):
    graph_filename: str = "cognee_graph.pkl"
    graph_database_provider: str = "NETWORKX"
    graph_database_url: str = ""
    graph_database_username: str = ""
    graph_database_password: str = ""
    graph_database_port: int = 123
    graph_file_path: str = os.path.join(relational_config.database_directory_path,graph_filename)
    graph_engine: object =  GraphDBType.NETWORKX
    graph_model: object = DefaultGraphModel

    model_config = SettingsConfigDict(env_file = ".env", extra = "allow")

    def to_dict(self) -> dict:
        return {
            "graph_filename": self.graph_filename,
            "graph_database_provider": self.graph_database_provider,
            "graph_topology": self.graph_topology,
            "graph_file_path": self.graph_file_path,
            "graph_database_url": self.graph_database_url,
            "graph_database_username": self.graph_database_username,
            "graph_database_password": self.graph_database_password,
            "graph_database_port": self.graph_database_port,
            "graph_engine": self.graph_engine

        }

@lru_cache
def get_graph_config():
    return GraphConfig()



""" This module is used to set the configuration of the system."""
from cognee.base_config import get_base_config
from cognee.infrastructure.databases.graph.config import get_graph_config
from cognee.infrastructure.data.chunking.config import get_chunk_config
from cognee.modules.cognify.config import get_cognify_config


cognify_config = get_cognify_config()
chunk_config = get_chunk_config()
graph_config = get_graph_config()
base_config = get_base_config()

class config():
    @staticmethod
    def system_root_directory(system_root_directory: str):
        base_config.system_root_directory = system_root_directory

    @staticmethod
    def data_root_directory(data_root_directory: str):
        base_config.data_root_directory = data_root_directory

    @staticmethod
    def monitoring_tool(monitoring_tool: object):
        base_config.monitoring_tool = monitoring_tool

    @staticmethod
    def set_classification_model(classification_model: object):
        cognify_config.classification_model =  classification_model

    @staticmethod
    def set_summarization_model(summarization_model: object):
        cognify_config.summarization_model=summarization_model


    @staticmethod
    def set_labeling_model(labeling_model: object):
        cognify_config.labeling_model =labeling_model


    @staticmethod
    def set_graph_model(graph_model: object):
        graph_config.graph_model =graph_model


    @staticmethod
    def set_cognitive_layer_model(cognitive_layer_model: object):
        cognify_config.cognitive_layer_model =cognitive_layer_model


    @staticmethod
    def set_graph_engine(graph_engine: object):
        graph_config.graph_engine =graph_engine

    @staticmethod
    def llm_provider(llm_provider: str):
        graph_config.llm_provider = llm_provider

    @staticmethod
    def llm_endpoint(llm_endpoint: str):
        graph_config.llm_endpoint = llm_endpoint

    @staticmethod
    def llm_model(llm_model: str):
        graph_config.llm_model = llm_model

    @staticmethod
    def intra_layer_score_treshold(intra_layer_score_treshold: str):
        cognify_config.intra_layer_score_treshold =intra_layer_score_treshold


    @staticmethod
    def connect_documents(connect_documents: bool):
        cognify_config.connect_documents = connect_documents

    @staticmethod
    def set_chunk_strategy(chunk_strategy: object):
        chunk_config.chunk_strategy = chunk_strategy


    @staticmethod
    def set_graph_topology(graph_topology: object):
        get_cognify_config.graph_topology =graph_topology




from typing import List, Dict, Any, Union, Optional

from cognee.infrastructure import infrastructure_config
from cognee.infrastructure.databases.graph.get_graph_client import get_graph_client

from cognee.modules.topology.topology import TopologyEngine, GitHubRepositoryModel
import pandas as pd
from pydantic import BaseModel


USER_ID = "default_user"

async def add_topology(directory: str = "example", model: BaseModel = GitHubRepositoryModel) -> Any:
    graph_db_type = infrastructure_config.get_config()["graph_engine"]

    graph_client = await get_graph_client(graph_db_type)

    graph_topology = infrastructure_config.get_config()["graph_topology"]

    engine = TopologyEngine()
    topology = await engine.infer_from_directory_structure(node_id=USER_ID, repository=directory, model=model)

    def flatten_model(model: BaseModel, parent_id: Optional[str] = None) -> Dict[str, Any]:
        """Flatten a single Pydantic model to a dictionary handling nested structures."""
        result = {**model.dict(), 'parent_id': parent_id}
        if hasattr(model, 'default_relationship') and model.default_relationship:
            result.update({
                'relationship_type': model.default_relationship.type,
                'relationship_source': model.default_relationship.source,
                'relationship_target': model.default_relationship.target
            })
        return result

    def recursive_flatten(items: Union[List[Any], BaseModel], parent_id: Optional[str] = None) -> List[Dict[str, Any]]:
        """Recursively flatten nested Pydantic models or lists of models."""
        if isinstance(items, list):
            return [entry for item in items for entry in recursive_flatten(item, parent_id)]
        elif isinstance(items, BaseModel):
            flat = [flatten_model(items, parent_id)]
            for field, value in items:
                if isinstance(value, (BaseModel, list)):
                    flat.extend(recursive_flatten(value, items.dict().get('node_id', None)))
            return flat
        else:
            return []

    def flatten_repository(repo_model: BaseModel) -> List[Dict[str, Any]]:
        """ Flatten the entire repository model, starting with the top-level model """
        return recursive_flatten(repo_model)

    flt_topology = flatten_repository(topology)

    df = pd.DataFrame(flt_topology)

    print(df.head(10))

    for _, row in df.iterrows():
        node_data = row.to_dict()
        node_id = node_data.pop('node_id')

        # Remove 'node_id' and get its value
        await graph_client.add_node(node_id, node_data)
        if pd.notna(row['relationship_source']) and pd.notna(row['relationship_target']):
            await graph_client.add_edge(row['relationship_source'], row['relationship_target'], relationship_name=row['relationship_type'])

    return graph_client.graph

if __name__ == "__main__":
    async def test() -> None:
        # Uncomment and modify the following lines as needed
        # await prune.prune_system()
        #
        # from cognee.api.v1.add import add
        # data_directory_path = os.path.abspath("../../../.data")
        # # print(data_directory_path)
        # # config.data_root_directory(data_directory_path)
        # # cognee_directory_path = os.path.abspath("../.cognee_system")
        # # config.system_root_directory(cognee_directory_path)
        #
        # await add("data://" + data_directory_path, "example")

        # graph = await add_topology()

        graph_db_type = infrastructure_config.get_config()["graph_engine"]

        graph_client = await get_graph_client(graph_db_type)
        #
        from cognee.utils import render_graph

        await render_graph(graph_client.graph, include_color=True, include_nodes=False, include_size=False)

    import asyncio
    asyncio.run(test())



# from cognee.shared.data_models import GraphDBType
#
# import logging
# logging.basicConfig(level=logging.DEBUG,
#                     format='%(asctime)s - %(levelname)s - %(message)s')
#
#
# import asyncio
# from cognee.shared.data_models import KnowledgeGraph, DefaultCognitiveLayer, TextContent, TextSubclass, DefaultContentPrediction
# from cognee.modules.cognify.llm.generate_graph import generate_graph
#
# required_layers_one = DefaultContentPrediction(
#     label=TextContent(
#         type='TEXT',
#         subclass=[TextSubclass.ARTICLES]
#     )
# )
#
# print(required_layers_one.dict())
#
#
# #
# #
# #
# # if __name__ == "__main__":
# #     async def main():
# #         logging.info("Hello")
# #         from cognee.infrastructure.databases.graph.get_graph_client import get_graph_client
# #
# #         client = await get_graph_client(graph_type=GraphDBType.NEO4J)
# #
# #         print(client)
# #
# #         await client.add_node("1", name="Alice")
# #
# #
# #         from cognee.utils import render_graph
# #         # graph_url = await render_graph(graph)
# #         # print(graph_url)
# #
# #
# #     asyncio.run(main())
import duckdb
# import networkx as nx
#
# # Create a new graph
# G = nx.Graph()
#
# # Add some nodes to the graph
# G.add_node(1)
# G.add_node(2)
# G.add_node(3)
#
# # Node ID you're looking for
# node_id = 2
#
# # Check if the node exists in the graph
# if G.has_node(node_id):
#     print(f"Node with ID {node_id} exists in the graph.")
# else:
#     print(f"Node with ID {node_id} does not exist in the graph.")

from nltk.parse.generate import generate, demo_grammar

from cognee.infrastructure import infrastructure_config

# from nltk import CFG
# grammar = CFG.fromstring("Utilizes the quantum superposition.")
# print(grammar)
# import nltk
# from nltk import CFG
# from nltk.tokenize import word_tokenize
# from nltk.parse import RecursiveDescentParser
#
# nltk.download('wordnet')
# from nltk.corpus import wordnet as wn
#
# # Define a CFG with parts of speech tags and catch-all rules for unknown words
# grammar = CFG.fromstring("""
#   S -> NP VP
#   NP -> Det N | Det N PP | N
#   VP -> V | V NP | V NP PP | V Adv
#   PP -> P NP
#   Det -> 'a' | 'the'
#   N -> 'dog' | 'park' | UNKNOWN_NOUN
#   V -> 'saw' | 'walked' | UNKNOWN_VERB
#   P -> 'in' | 'on'
#   Adv -> 'quickly' | UNKNOWN_ADV
#   """)
#
# # Initialize a recursive descent parser with the grammar
# parser = RecursiveDescentParser(grammar)
#
# def get_wordnet_pos(treebank_tag):
#     """Converts treebank POS tags to WordNet POS tags."""
#     if treebank_tag.startswith('J'):
#         return wn.ADJ
#     elif treebank_tag.startswith('V'):
#         return wn.VERB
#     elif treebank_tag.startswith('N'):
#         return wn.NOUN
#     elif treebank_tag.startswith('R'):
#         return wn.ADV
#     else:
#         return None
#
# def get_unknown_pos(word):
#     """Finds the most likely POS tag for an unknown word using WordNet."""
#     synsets = wn.synsets(word)
#     if not synsets:
#         return 'UNKNOWN_NOUN'  # Default to NOUN if word not found
#     pos_counts = [get_wordnet_pos(s.pos()) for s in synsets]
#     most_common_pos = max(set(pos_counts), key=pos_counts.count)
#     if most_common_pos == wn.NOUN:
#         return 'UNKNOWN_NOUN'
#     elif most_common_pos == wn.VERB:
#         return 'UNKNOWN_VERB'
#     elif most_common_pos == wn.ADJ:
#         return 'UNKNOWN_ADJ'
#     elif most_common_pos == wn.ADV:
#         return 'UNKNOWN_ADV'
#     else:
#         return 'UNKNOWN_NOUN'
#
# def preprocess_sentence(sentence, original_grammar):
#     """Dynamically adjusts the grammar for unknown words in the sentence."""
#     tokens = word_tokenize(sentence)
#     new_productions = original_grammar.productions()
#
#     for word in tokens:
#         if not any(word in str(prod) for prod in original_grammar.productions()):
#             unknown_pos = get_unknown_pos(word)
#             # Create a new production for the unknown word and add it to the list of productions
#             new_production = CFG.fromstring(f"{unknown_pos} -> '{word}'").productions()[0]
#             new_productions.append(new_production)
#
#     # Create a new grammar with the adjusted set of productions
#     adjusted_grammar = CFG(original_grammar.start(), new_productions)
#     return tokens, adjusted_grammar
#
#
# # Example sentence
# sentence = "Alice utilizes quantum superposition"
#
# # Preprocess the sentence and adjust the grammar for unknown words
# tokens, adjusted_grammar = preprocess_sentence(sentence, grammar)
#
# print(tokens)
# print(adjusted_grammar)

# Parse the preprocessed sentence using the adjusted grammar
# parser = RecursiveDescentParser(adjusted_grammar)
#
# for tree in parser.parse(tokens):
#     print(tree)
# import nltk
# from nltk.tokenize import word_tokenize
# from nltk.tag import pos_tag
# from nltk.chunk import ne_chunk
#
# # Ensure that the necessary NLTK resources are downloaded
# nltk.download('maxent_ne_chunker')
# nltk.download('words')
#
# # The sentence you want to tag and recognize entities in
# sentence = "Apple Inc. is an American multinational technology company headquartered in Cupertino, California."
#
# # Tokenize the sentence into words
# tokens = word_tokenize(sentence)
#
# # Perform POS tagging on the tokenized sentence
# tagged = pos_tag(tokens)
# print(tagged)
#
# # Perform Named Entity Recognition (NER) on the tagged tokens
# entities = ne_chunk(tagged)
#
# # Print the named entities
# print(entities)
#
# # Alternatively, to display the entities more clearly, you can use:
# entities.pprint()
#
# infra_config = infrastructure_config.get_config()
# # db = duckdb.connect(infra_config["database_path"])
#
# db_engine = infra_config["database_engine"]
#
# data = [
#     {
#         'document_id': 'doc1',
#         'layer_id': 'layer1',
#         'created_at': '2022-01-01 00:00:00',
#         'updated_at': '2022-01-01 00:00:00'
#     },
#     {
#         'document_id': 'doc2',
#         'layer_id': 'layer2',
#         'created_at': '2022-01-02 00:00:00',
#         'updated_at': '2022-01-02 00:00:00'
#     }
# ]
#
# # Call the load_cognify_data method
# db_engine.load_cognify_data(data)
#
# # Call the get_cognify_data method and print the results
# cognify_data = db_engine.get_cognify_data()
# print(cognify_data)



import networkx as nx

# Create an empty graph
G = nx.Graph()

# Add nodes with parameters as dictionaries
G.add_node(1, params={'color': 'red', 'size': 10, 'nested_params': {'shape': 'circle', 'opacity': 0.5}})
G.add_node(2, params={'color': 'blue', 'size': 15})
G.add_node(3, params={'color': 'green', 'size': 20})

# Add edges
G.add_edge(1, 2)
G.add_edge(2, 3)

# Print nodes with parameters
for node, params in G.nodes(data='params'):
    print("Node:", node)
    print("Parameters:", params)

# Print edges
print("Edges:", G.edges())





import os
from functools import lru_cache
from pydantic_settings import BaseSettings, SettingsConfigDict
from cognee.infrastructure.databases.relational import DuckDBAdapter
from cognee.base_config import get_base_config
config = get_base_config()

class RelationalConfig(BaseSettings):
    db_path: str =  os.path.join(config.system_root_directory, "databases")
    db_name: str =  "cognee.db"
    db_host: str =  "localhost"
    db_port: str =  "5432"
    db_user: str = "cognee"
    db_password: str =  "cognee"
    db_engine: object = DuckDBAdapter(
        db_name=db_name,
        db_path=db_path
    )
    database_engine: object = db_engine
    db_file_path:str = os.path.join(db_path, db_name)
    database_path: str = os.path.join(config.system_root_directory, "databases")
    database_directory_path: str = db_path

    model_config = SettingsConfigDict(env_file = ".env", extra = "allow")

    def to_dict(self) -> dict:
        return {
            "db_path": self.db_path,
            "db_name": self.db_name,
            "db_host": self.db_host,
            "db_port": self.db_port,
            "db_user": self.db_user,
            "db_password": self.db_password,
            "db_engine": self.db_engine,
            "database_path": self.database_path,
        }

@lru_cache
def get_relationaldb_config():
    return RelationalConfig()


import os
import glob
from pydantic import BaseModel, Field
from typing import Dict, List, Optional, Union, Type, Any, Tuple
from datetime import datetime

from cognee import config
from cognee.base_config import get_base_config
from cognee.infrastructure import infrastructure_config
from cognee.modules.cognify.config import get_cognify_config
from cognee.modules.topology.infer_data_topology import infer_data_topology
cognify_config = get_cognify_config()
base_config = get_base_config()

class Relationship(BaseModel):
    type: str = Field(..., description="The type of relationship, e.g., 'belongs_to'.")
    source: Optional[str] = Field(None, description="The identifier of the source id of in the relationship being a directory or subdirectory")
    target: Optional[str] = Field(None, description="The identifier of the target id in the relationship being the directory, subdirectory or file")
    properties: Optional[Dict[str, Any]] = Field(None, description="A dictionary of additional properties and values related to the relationship.")


class Document(BaseModel):
    node_id: str
    title: str
    description: Optional[str] = None
    default_relationship: Relationship


class DirectoryModel(BaseModel):
    node_id: str
    path: str
    summary: str
    documents: List[Document] = []
    subdirectories: List['DirectoryModel'] = []
    default_relationship: Relationship


DirectoryModel.update_forward_refs()


class DirMetadata(BaseModel):
    node_id: str
    summary: str
    owner: str
    description: Optional[str] = None
    directories: List[DirectoryModel] = []
    documents: List[Document] = []
    default_relationship: Relationship


class GitHubRepositoryModel(BaseModel):
    node_id: str
    metadata: DirMetadata
    root_directory: DirectoryModel


class TopologyEngine:
    def __init__(self) -> None:
        self.models: Dict[str, Type[BaseModel]] = {}

    async def populate_model(self, directory_path: str, file_structure: Dict[str, Union[Dict, Tuple[str, ...]]], parent_id: Optional[str] = None) -> DirectoryModel:
        directory_id = os.path.basename(directory_path) or "root"
        directory = DirectoryModel(
            node_id=directory_id,
            path=directory_path,
            summary=f"Contents of {directory_id}",
            default_relationship=Relationship(type="contains", source=parent_id, target=directory_id)
        )

        for key, value in file_structure.items():
            if isinstance(value, dict):
                # Recurse into subdirectory
                subdirectory_path = os.path.join(directory_path, key)
                subdirectory = await self.populate_model(subdirectory_path, value, parent_id=directory_id)
                directory.subdirectories.append(subdirectory)
            elif isinstance(value, tuple) and value[0] == 'file':
                # Handle file
                document = Document(
                    node_id=key,
                    title=key,
                    default_relationship=Relationship(type="contained_by", source=key, target=directory_id)
                )
                directory.documents.append(document)

        return directory

    async def infer_from_directory_structure(self, node_id: str, repository: str, model: Type[BaseModel]) -> GitHubRepositoryModel:
        """ Infer the topology of a repository from its file structure """

        path = base_config.data_root_directory
        path = path + "/" + str(repository)
        print(path)

        if not os.path.exists(path):
            raise FileNotFoundError(f"No such directory: {path}")

        root: Dict[str, Union[Dict, Tuple[str, ...]]] = {}
        for filename in glob.glob(f"{path}/**", recursive=True):
            parts = os.path.relpath(filename, start=path).split(os.path.sep)
            current = root
            for part in parts[:-1]:  # Traverse/create to the last directory
                if part not in current:
                    current[part] = {}
                current = current[part]
            last_part = parts[-1]
            if os.path.isfile(filename):
                current[last_part] = ("file", ...)  # Placeholder for file content or metadata
            elif os.path.isdir(filename):
                if last_part not in current:  # Only create a new directory entry if it doesn't exist
                    current[last_part] = {}

        root_directory = await self.populate_model('/', root)

        repository_metadata = DirMetadata(
            node_id="repo1",
            summary="Example repository",
            owner="user1",
            directories=[root_directory],
            documents=[],
            default_relationship=Relationship(type="contained_by", source="repo1", target=node_id)
        )

        active_model = GitHubRepositoryModel(
            node_id=node_id,
            metadata=repository_metadata,
            root_directory=root_directory
        )

        return active_model

    def load(self, model_name: str) -> Optional[Type[BaseModel]]:
        return self.models.get(model_name)

    def extrapolate(self, model_name: str) -> None:
        # This method would be implementation-specific depending on what "extrapolate" means
        pass


if __name__ == "__main__":
    data_directory_path = os.path.abspath("../../../.data")
    print(data_directory_path)
    config.data_root_directory(data_directory_path)
    cognee_directory_path = os.path.abspath("../.cognee_system")
    config.system_root_directory(cognee_directory_path)

    async def main() -> None:
        engine = TopologyEngine()
        # model = engine.load("GitHubRepositoryModel")
        # if model is None:
        #     raise ValueError("Model not found")
        result = await engine.infer_from_directory_structure("example_node_id", "example_repo", GitHubRepositoryModel)
        print(result)

    import asyncio
    asyncio.run(main())
    # result = engine.extrapolate("GitHubRepositoryModel")
    # print(result)



from functools import lru_cache
from pydantic_settings import BaseSettings, SettingsConfigDict

class LLMConfig(BaseSettings):
    llm_provider: str = "openai"
    llm_model: str = "gpt-4o"
    llm_endpoint: str = ""
    llm_api_key: str = ""
    llm_streaming:bool = False

    model_config = SettingsConfigDict(env_file = ".env", extra = "allow")

    def to_dict(self) -> dict:
        return {
            "provider": self.llm_provider,
            "model": self.llm_model,
            "endpoint": self.llm_endpoint,
            "apiKey": self.llm_api_key,
        }

@lru_cache
def get_llm_config():
    return LLMConfig()
